# Data Structures

**Data Structures** ~ These are organized mechanisms for efficiently storing, accessing, and manipulating data. They are reusable building blocks for solving complex problems. They should be clearly designed, generalized, implemented with reusability, extensibility, and scalability.  

**Type** ~ This is a collection of values. Simple versions of these include int and bool. Aggregate and more complex versions of these are known as **classes**. %%A boolean is either true or false%%.
**Data Type** ~ This is a type with operations on it. 
**Abstract Data Types** ~ These are data types that are independent of, and must not have, implementation. They are similar to Java interfaces, where all methods are abstract.

^^^
**Algebraic Specification** ~ This is an outline of an abstract data type. It includes what other abstract data types it uses, what it defines, its operations, its axioms, its preconditions, and more. &&**Operations** specify transformations of the data&& &&**Axioms** specify rules that must always be satisfied by the implementations of our operations&& &&**Partial Operations**:When an operation is limited by pre-conditions, then we say it is a partial operation and use the -/-> notation.&&

%%
```
adt Variable
    uses Any
    defines Variable<T: Any>
    operations:
        new: T ---> Variable<T>
        put: Variable<T> x T ---> Variable<T>
        get: Variable<T> --> T
    axioms:
        get(new(t)) = t
        get(put(v, t)) = t
    
```
%%
^^^

^^^
**Array** ~ This is a data structure that stores a list of data. The size of the list cannot change, and elements can be modified.
%%
```
adt Array
    uses Any, Integer
    defines Array<T: Any>
    operations:
        new: Integer x T -/-> Array<T>
        get: Array<T> x Integer -/-> T
        put: Array<T> x Integer x T -/-> Array<T> 
        length: Array<T> ---> Integer
    preconditions:
        new(l, t): 0 < l
        get(a, i): 0 <= i < length(a)
        put(a, i, t): 0 <= i < length(a)
    axioms:
        get(new(n, t), i) = t
        get(put(a, i, t), j) = (if i = j then t else get(a, j))
        length(new(n, t)) = n
        length(put(a, i, t)) = length(a)
```
%%
^^^

^^^
**Linked List** ~ This is a data structure consisting of Nodes which store a value and a pointer to a next Node.

^^^

^^^
**Stack** ~ This is a data structure that adds to and deletes from the same end. It embodies the Last In, First Out (LIFO) concept, and consists of four methods: 1. push 2. pop 3. top/peek 4. empty. 
%%
```
adt Stack
    uses Any, Boolean
    defines Stack<T: Any>
    operations:
        new: ---> Stack<T>
        push: Stack<T> x T ---> Stack<T>
        pop: Stack<T> -/-> Stack<T>
        top: Stack<T> -/-> T
        empty: Stack<T> ---> Boolean
    preconditions:
        pop(s): not empty(s)
        top(s): not empty(s)
    axioms:
        empty(new()
        not empty(push(s, t))
        top(push(s, t)) = t
        pop(push(s, t)) =
```
%%
^^^

**Amortized Analsis** ~ This is the calculation of the \_\_\_ cost per operation for a sequence of n operations, which is equal to the total cost of the operations divided by n.

^^^
**Queue** ~ This is a data structure that adds to one end and deletes from the other.
%%
```
adt Queue
    uses
    defines
    operations:
        new:
        enqueue:
        dequeue:
        front:
        empty:
    preconditions:
    axioms:
```
%%
^^^

^^^
**Deque** ~ This is a data structure that adds to or deletes from either end. &&It stands for double-ended queue. It is both a stack and a queue.&&
%
```
adt Deque
    uses Any
    defines Deque
    operations:
        new:
        length:
        empty:
        front:
        back:
        insertFront:
        removeFront:
        insertBack:
        removeBack:
    preconditions:
    axioms:
```
%
^^^

^^^
**(Positional) List** ~ This is a collection of positions with relative ordering. The relative position is what matters, not the value itself.
%%
```
adt Position
    uses Any
    defines Position
    operations:
        get:
        put:
    preconditions:
    axioms:
```
%%

%%
```
adt List
    uses Position, Boolean
    defines List
    operations:
        new:
        length:
        empty:
        insertFront:
        insertBack:
        insertBefore:
        insertAfter:
        remove:
        removeFront:
        removeBack:
        front:
        back:
        previous:
        next:
        first:
        last:
        iterator:
    preconditions:
    axioms:
```
%%
^^^

^^^
**Set** ~ This is a collection of objects with position relative to values. The values of the data being stored determine where and how it gets stored. &&A version of the Set called OrderedSet exists and it has the same operations as Set, but the elements are expected to be iterated over in order based on their value.&&
```
adt Set
    uses: Boolean, Integer, Any
    defines Set
    operations:
        new: ---> Set<T>
        has: Set<T> x T ---> Boolean
        insert: Set<T> x T ---> Set<T>
        remove: Set<T> x T ---> Set<T>
        size: Set<T> ---> Integer
        iterator:
        union:
        intersect:
        subtract:
    axioms:
        not has(new(), t)
        has(insert(s, t), x) = if t=x then true else has(s, x)
        has(remove(s, t), x) = if t=x then
```
^^^

^^^
**(Rooted) Tree** ~ This is a position based data structure in which positions have parents and each position has an unlimited number of children. Connections between positions (nodes) are called **edges**. The parent of all nodes is called the **root**. A \_\_\_ of a child is called a **sub\_\_\_**. Nodes without any children are called **leaves**. The **depth** of the root is 0 and counts how many "ancestors" (parents & parents of parents) a node has. The **height** of any leaf is 0 and counts the maximum number of descendents (children and children of children) a node has. The **path** is the number of edges between two nodes.
%%
```
```
%%
^^^

**Tree Traversals** ~ These are different ways to iterate through a tree. They include breadth-first (level order) search, and depth-first search (includes: pre-order, post-order, in-order). Breadth-first starts with the root, then the children of the root, then the children of the children. Depth-first pre-order visits the root of a subtree, then recursively do a pre-order traversal of each child. Depth-first post-order does a post-order traversal of each child, then visit root of subtree. Depth-first in-order (for binary trees only) recursively does in-order traversal of the left child, then visits the subtree root, and then does in-order traversal of the right child.

^^^
**Priority Queue** ~ This is a variation of a standard queue - values go in but only the "best" come out, one at a time. These are usually minimum or maximum based. 
%%
```
adt PriorityQueue
    uses:
    defines: PriorityQueue
    operations:
        new:
        best: 
        remove:
        insert:
        empty:
    preconditions:
    axioms:
```
%%
^^^

^^^
**Binary Heap** ~ This is a complete binary tree that is filled on all levels, except the last, and is filled from left to right. The structure is ordered in that each node value is greater than or equal to its children's values, so that the minimum/maximum is always at the root.
%%
```
adt BinaryHeap
    uses:
    defines: BinaryHeap
    operations:
        insert:
        remove:
```
%%
^^^

^^^
**Map / Associated Arrays** ~ This is a collection of key, value pairs called entries/entities. Keys must be unique and immutable. &&Maps that have duplicated keys are called **dictionaries** (hashmaps). If you sort your keys (are comparable), then you have an **ordered map** (binary search tree (BST), balanced BST, AVL tree, Red/black, 2-3, B-tree, Treap, etc.). Likewise, you can have an **ordered dictionary**.&&
%%
```
adt Map
    uses:
    defines: Map
    operations:
        size:
        put:
        get:
        has:
        insert:
        remove:
        
        (Dictionary only)
        findAll:
        removeAll:
    iterators:
        keys:
        values:
        entries:
```
%%
^^^

^^^
**Binary Search Tree** ~ This is a binary tree where each node has 0, 1, or 2 children only. It has complete ordering - everything in the left of the subtree of the parent is less than than parent, everything in the right of the subtree of the parent is greater than the parent, and if duplicates are allowed, they must be put consistently on the same side. This allows for an in-order traversal that gives the nodes in ascending order. It's design is inspired by binary search, and that is what it uses to find a node.
%%
```
adt BST:
    uses:
    defines: BST
    operations:
        find:
        insert:
        remove:
```
%%
^^^

^^^
**AVL Tree / Balanced Binary Search Tree** ~ This tree not only has an order property, but also a balance property. For every internal position p of this tree, the heights of the children of p differ by at most 1. The height of this binary tree that satisfies the balance property is O(log N) for N nodes total. This data structure stores the height of each node and adds post-processing to a binary tree in order to maintain the balance property. 

Insertion: 
(1) Insert item as in binary search tree. 
(2) Starting with new node, go up to update heights.
(3) If find a node that has become unbalanced, do rotation to rebalance.
    Rebalance reduces height of subtree by 1, so no need to propegate up - only one adjustment needed at most because it was previously balanced before the insert.
O(log(n)) to insert and O(1) rotation to rebalance.

Removing:
(1) Remove as in binary search tree.
(2) Ancestor of actual deleted leaf node (not value replaced node) may become unbalanced.
(3) Go up from deleted node to find first unbalanced ancestor.
(4) Do rotation : may reduce the height of the sub-tree, causing further unbalances for ancestors.
(5) O(log(N)) to balance and remove.

Rotating:
(1) Singles (same direction) and doubles (opposite directions)
(2) How to rotate depends on the balance factor of the (sub)root and the child on the longer side.
(3) Remember that we rebalance from the bottom up, so both children of an unbalanced node must have -1 <= balance factor <= 1.
%%
```
adt AVLTree:
    uses: TreeNode
    defines: AVLTree
    operations:
        insert:
        remove:
```
%%
^^^

^^^
**Treap** ~ This is an ordered map in which each key-value pair is assigned a random priority. The binary search order property is based on keys and the priorities are used to create a heap. This is a **probabilistic** map implementation, and there is a good *expected* time complexity O(log(n)), but not a guaranteed one. O(N) worst case, but unlikely.

Find: O(height)
    Find as if in BST
Insertion: O(height)
    Insert each node as if in BST
    Balance according to Heap priorities & order should be maintained automatically
Remove: O(height)
    Remove as if in BST
    Balance according to Heap priorities & order should be maintained automatically
    
Benefits:
    Easier rotations than AVLTree (only singles, directions easy to determine)
    No need to track heights
    Good performance in practice
%%
```
adt Treap:
    uses:
    defines: Treap
    operations:
        has:
        find:
        insert:
        
```
%%
^^^

^^^
**Hash Map** ~ This is a data structure used for unordered maps and dictionaries. This has a worst case O(N) access/item, but the average case is O(1) (constant) time. It is an implementation of a Map and converts keys to ais an implementation of a Map and converts keys to array indices through hashing. 

`hashCode()` is defined in the Object class and by default returns the integer representation of the memory address. This is not ideal because if you have two objects that you consider identical, Java may not consider them identical because they are two separate objects.

The three C's of hashing:
Code
Compress 
Collisions

%%
```
adt HashMap:
    uses:
    defines: HashMap
    operations:
        hash:
        has:
        get:
        find:
        insert:
        put:
        remove:
```
%%
^^^

^^^
**Graph** ~ This is a mathematical structure for representing relationships. It consists of a set of **nodes/vertices** connected by **edges/arcs**.**edges/arcs**. There are two types of these - directed and undirected. The **directed** ones have direction associated with edges (uni/bi-directional), where as **undirected** only have bidirectional edges. Each edge has one or two vertices associated with it, called its endpoints (an edge is said to connect endpoints). Two vertices are **adjacent** if they are attached directly by at least one edge. They are **connected** if they are connected by a sequence of edges. The **neighborhood** of a vertex is the set of all vertices in G that are adjacent to a vertex v, denoted N(v). An **adjacency matrix** is a representation of this as a 2-dimensional arrayas a 2-dimensional array, using boolean 0 and 1's to depict what vertices are connected. The **degree** of a vertex is the number of edges incident with v, and loops are counted as two. The **in-degree** of a vertex v is the number of edges that terminate at v. The **out-degree** is the number coming from v. A This structure is **complete** if it connects all nodes together directly. The **path** from one node to another is the sequence of edges, and the **length** is the number of edges. This is called **connected** if there is a path between every pair of vertices. A **cycle** is a path that starts and ends at the same node. A **simple** path or cycle does not repeat any nodes or edges. 

&&A tree is a connected acyclic graph. A rooted tree is a tree with a special node that is connected to all other nodes.
&& 

&&General graph search:
Input : Graph G = (V, E) and a starting vertex s in V
Goal : Identify the vertices of V reachable from s in G

```
dfs:
    mark s as explored, all other vertices as unexplored
    S := stack
    while S is not empty:
        pop top vertex v
        if v is unexplored
            mark v as explored
                for each edge (v, w) in v's adjacency list:
                    if w is unexplored:
                        push w to front of S
```
&&

%%
For a graph G = (V, E) with vertex set V and edge set E. N = |V| and M = |E|. In a simple, connected, undirected graph, M is bounded by O(N) and O(N^2).
```
adt Graph:
    uses: : 
    defines: Graph
    operations:
        insertVertex(v)
        removeVertex(v)
        inse
        removeVertex(v)
        insertEdge(u, v)
        removeEdge(e)
        numVertices()
        numEdges()
        vertices()
        edges()
        areAdjacent(u, v)
        isIncident(v, e)
        degree(v)
        neighbors(v)
```
%%
^^^

# Algorithms

## Analysis of Algorithms

**Algorithms** ~ These are sequences of steps/actions that produce desired outputs, potentially given specific inputs.
**Scientific Method** ~ This is our framework for predicting performance and comparing algorithms. (1) Observe our algorithm. (2) Hypothesize a model that is consistent with our observations. (3) Predict events using the hypothesis. (4) Verify the predictions by making further observations. (5) Validate by repeating until the hypothesis and observations agree. -- Experiments must be reproducible and Hypotheses must be falsifiable.

**Empirical Analysis** ~ This is an analysis of an algorithm by executing the program to perform experiments. Creating a model by assuming the time it takes for an algorithm to run, this model enables us to make predictions.
**Mathematical Analysis** ~ This is an analysis of an algorithm by counting the frequency of operations. We often use tilde notations to simplify analysis, allowing us to explain the behavior through our model.

**Cost Model** ~ This is a way of finding out the total running time of an algorithm. It involves using some basic operation (usually the one with highest cost and frequency) as a proxy for the running time.

!**Order-of-growth Classifications** ~ These are the set of functions that describe the time efficiency of typical algorithms. They include: 1, log N, N, N log N, N^2, N^3, and 2^N. Generally, only linearithmic and more efficient algorithms are of practical use. ![](https://i.imgur.com/NirOO08.png)

!**Types of Algorithm Analyses** ~ These include (1) Best case: The lower bound on the cost of the algorithm, determined by the "easiest" input. This is the goal cost for all inputs. (2) Worst case: The upper bound on the cost, determined by the "most difficult" input. This is a guaranteed cost for all inputs. (3) Average case: The expected cost for random input. This requires a model for "random" input, but provides a way to predict performance. 

**Theory of Algorithms** ~ The goals of this theory is to establish the "difficulty" of problems and develop "optimal" algorithms. The approach includes suppression of details in analysis and elimination of variability in the input model by focusing on the worst case.

**Common Notations in Theory of Algorithms** ~ These include (1) **Tilde Notation** ~n: This involves ignoring lower order terms because as the size of input gets high, these lower order terms become negligible, providing an approximate bound on an algorithm. (2) **Big Theta** Θ(n), which provides the asymptotic order of growth and is used to classify algorithms. (3) **Big O** O(n), which provides the upper bounds of an algorithm. (4) **Big Omega** Ω(n), which is used to develop lower bounds.

**Shallow Memory Usage** ~ This refers to the total memory usage for a data type without referenced objects.
**Deep Memory Usage** ~ This refers to the total memory usage for a data type including references, adding memory recursively for each referenced object. 

## Pure Algorithms

### Sorting Algorithms
**Bubble Sort** ~
**Insertion Sort** ~ 
**Selection Sort** ~
**Selection Sort** ~ 
**Heap Sort** ~ 
**Quick Sort** ~ This is a sort that picks a pivot element from alongthe values apicks a pivot element from alongthe values and then splits the collection by comparing each value to pivot (put those < pivot on left, those >= pivot on right). Then, recursively sort each group. Best case: O(n log n) if splits close to half. Worst case: O(n^2) if splits only one each time. Average case: O(n log n)
**Merge Sort** ~ 
**Bucket Sort** ~
**Radix Sort** ~  This is a sort using d-tuples, with M possible values for each element in the tuple. You start with the back-most (right-most) dimension and do bucket sort on each dimension, pulling values out in FIFO order. Runtime: O()
**Stable Sorting** ~ This is a concept in sorting that maintains the order of elements with duplicate values. This is important if the sort order is based on a subset of data members from a larger object. Bucket and radix sort are stable if you us




## Applied Algorithms
### Union Find

+++++
**The Problem of Dynamic Connectivity** ~
    * Given a set of N objects, a union command means to connect two objects. Is there a path connecting the two objects? Objects are connected if they are reflexive, symmetric, and transitive. 
    * Goal: Design a data structure for union-find:
        * Number of objects N can be huge
        * Number of operations M can be huge
        * Find queries and union commands may be intermixed.

+++++

+++++
!^**Union Find Eager Approach** 
* Data Structure:
    * Using an integer array id[] of size N, p and q are connected if and only if they have the same id. 

* Find:
    * Check if p and q have the same id to find out if two objects are connected. 

* Union:
    * To merge components containing p and q, change all entries whose id equals id[p] to id[q] ![](https://i.imgur.com/dQs6143.png). ![The elements at index 0, 1, 2, 5, 6, and 7 all have the same id, and are thus connected.]
~
```java
public class QuickFindUF {
    private int[] id;
    
    public QuickFindUF(int N) {
        id = new int[N];
        for (int i = 0; i < N.length; i++)
            id[i] = i;
    }
    
    public boolean connected(int p, int q) {
        return id[p] == id[q];
    }
    
    public void union(int p, int q) {
        int pid = id[p];
        int qid = id[q];
        for (int i = 0; i < id.length; i++)
            if (id[i] == pid)
                id[i] = qid;
    }
}
```
* Cost Model:
    * Initialize is O(n)
    * Find is O(1)
    * The union method is O(n) and we have to use it n times (for each element in id). This is too slow because n can be huge and n^2 does not scale well

+++++

+++++
^**Union Find Quick Union Approach**
* Data Structure:
    * Using an integer array id[] of size N, let id[i] be the parent of i. Thus the root of i is id[id[id[...id[i]...]]].

* Find:
    * Check if p and q have the same root.

* Union:
    * To merge components containing p and q, set the id of p's root to the id of q's root. ![](https://i.imgur.com/9Qn6EP5.png) 
    ![Without the connection, the root of 3, 2, and 4 is 9, and the root of 5 is 6. After the union beween p = 3 and q = 5, the roots 9 and 6 are connected. This is done by setting the id at 9 equal to the 6.]
~
```java
public class QuickUnionUF {
    private int[] id;
    
    public QuickUnionUF(int N) {
        id = new int[N];
    for (int i = 0; i < N; i++)
        id[i] = i;
    }
    
    private int root(int i) {
        while (i != id[i])
            i = id[i];
        return i;
    }
    
    public boolean connected(int p, int q) {
        return root(p) == root(q);
    }
    
    public void union(int p, int q) {
        int i = root(p);
        int j = root(q);
        id[i] = j;
    }
}
```

* Cost Model:
    * Trees can get too tall.
    * Find is expensive because it is O(n).
    * Union depends on the find method and worst case it is O(n).

+++++
^**Union Find Weighted Quick-Union Approach**
* Data Structure:
    * Using an integer array id[] of size N, let id[i] be the parent of i. Thus the root of i is id[id[id[...id[i]...]]]. Maintain an extra array sz[i] to count the number of objects in the tree rooted at i.

* Find:
    * Check if p and q have the same root.

* Union:
    * To merge components containing p and q, set the id of p's root to the id of q's root.
~
```java
public class WeightedQuickUnionUF {
    private int[] id;
    private int[] sz;
    
    public WeightedQuickUnionUF(int N) {
        id = new int[N];
        sz = new int[N];
        for (int i = 0; i < N; i++)
            id[i] = i;
    }
    
    private int root(int i) {
        while (i != id[i])
            i = id[i];
        return i;
    }
    
    public boolean connected(int p, int q) {
        return root(p) == root(q);
    }
    
    public void union(int p, int q) {
        int i = root(p);
        int j = root(q);
        if (i == j)
            return;
        if (sz[i] < sz[j]) {
            id[i] = j; 
            sz[j] += sz[i];
        } else {
            id[j] = i;
            sz[i] += sz[j];
        }
    }
}
```
* Running time:
    * The depth of any node x is at most lg N. (base 2)
        * Why? When does the depth of x increase?
            * The size of the tree containing x at least doubles each time.
            * This is because if we have two trees T1 and T2, these two trees will only merge when |T2| >= |T1|. 
            * The size of the tree containing x can only double at most lg(N) times.
            * Otherwise the smaller tree T1 will just go under the bigger tree T2 and the depth of x will not increase.
    * Find: Takes time proportional to the depth of p and q. O(lg(n))
    * Union: Takes constant time, given roots. O(lg(n))

+++++

+++++
^**Union Find Weighted Quick-Union Path Compression Approach**
* Data Structure:
    * Using an integer array id[] of size N, let id[i] be the parent of i. Thus the root of i is id[id[id[...id[i]...]]]. Maintain an extra array sz[i] to count the number of objects in the tree rooted at i.

* Find:
    * Check if p and q have the same root. On the way, make every other node in the path point to its grandparent (thereby halving the path length). 

* Union:
    * To merge components containing p and q, set the id of p's root to the id of q's root. Before doing so, identify the smaller tree because we want to link the root of the smaller tree to the root of the largest tree. Update the sz[] array.
~
```java
public class WeightedQuickUnionPCUF {
    private int[] id;
    private int[] sz;
    
    public WeightedQuickUnionPCUF(int N) {
        id = new int[N];
        sz = new int[N];
        for (int i = 0; i < N; i++)
            id[i] = i;
    }
    
    private int root(int i) {
        while (i != id[i]) {
            i = id[i];
            id[i] = id[id[i]];
        }
        return i;
    }
    
    public boolean connected(int p, int q) {
        return root(p) == root(q);
    }
    
    public void union(int p, int q) {
        int i = root(p);
        int j = root(q);
        if (i == j)
            return;
        if (sz[i] < sz[j]) {
            id[i] = j; 
            sz[j] += sz[i];
        } else {
            id[j] = i;
            sz[i] += sz[j];
        }
    }
}
```
* Running time:
    * Starting from an empty data structure, any sequence of M union-find ops on N objects makes <= c(N + M lg* N) array accesses. (lg* is the iterative star log function, meaning the number of times you need to take lg to get to 1, which is practically less than or equal to 5.)
        * Analysis can be improved to N + M \(\alpha\)(M, N).
    
+++++

**Percolation** ~ This a state of a system that is an N by N grid of sites where each site is open with probability p (or blocked with probability 1 - p) such that the top and bottom are connected by open sites. <e>In other words, if an open site is modeled by a white square and a blocked site is modeled by a black square, this is whether there is a path of white squares from the top to the bottom. <z>
**Monte Carlo Simulation** ~ This is a simulation where you initialize an N by N grid as completely blocked. Then we declare random sites open until the top is connected to the bottom and the system is percolated. This simulation is used to estimate the value for p where the chance of a system percolating shifts from likely to percolate to unlikely to percolate.

